"""
Nested filtering backend.
"""

from elasticsearch_dsl.query import Q
from django.core.exceptions import ImproperlyConfigured
from django_elasticsearch_dsl import fields

from six import string_types

from ...constants import (
    ALL_LOOKUP_FILTERS_AND_QUERIES,
    LOOKUP_FILTER_TERMS,
)

from ...compat import coreapi
from ...compat import coreschema
from .common import FilteringFilterBackend

__title__ = 'django_elasticsearch_dsl_drf.filter_backends.filtering.nested'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = ('NestedFilteringFilterBackend',)


class NestedFilteringFilterBackend(FilteringFilterBackend):
    """Nested filter backend.

    Example:

        >>> from django_elasticsearch_dsl_drf.constants import (
        >>>     LOOKUP_FILTER_TERM,
        >>>     LOOKUP_FILTER_PREFIX,
        >>>     LOOKUP_FILTER_WILDCARD,
        >>>     LOOKUP_QUERY_EXCLUDE,
        >>>     LOOKUP_QUERY_ISNULL,
        >>> )
        >>> from django_elasticsearch_dsl_drf.filter_backends import (
        >>>     NestedFilteringFilterBackend
        >>> )
        >>> from django_elasticsearch_dsl_drf.viewsets import (
        >>>     BaseDocumentViewSet,
        >>> )
        >>>
        >>> # Local article document definition
        >>> from .documents import ArticleDocument
        >>>
        >>> # Local article document serializer
        >>> from .serializers import ArticleDocumentSerializer
        >>>
        >>> class ArticleDocumentView(BaseDocumentViewSet):
        >>>
        >>>     document = ArticleDocument
        >>>     serializer_class = ArticleDocumentSerializer
        >>>     filter_backends = [NestedFilteringFilterBackend,]
        >>>     nested_filter_fields = {
        >>>         'country': {
        >>>             'field': 'continent.country.name.raw',
        >>>             'path': 'continent.country',
        >>>             'lookups': [
        >>>                 LOOKUP_FILTER_TERM,
        >>>                 LOOKUP_FILTER_TERMS,
        >>>                 LOOKUP_FILTER_PREFIX,
        >>>                 LOOKUP_FILTER_WILDCARD,
        >>>                 LOOKUP_QUERY_EXCLUDE,
        >>>                 LOOKUP_QUERY_ISNULL,
        >>>             ],
        >>>         }
        >>> }
    """

    @classmethod
    def prepare_filter_fields(cls, view):
        """Prepare filter fields.

        :param view:
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Filtering options.
        :rtype: dict
        """
        if not hasattr(view, 'nested_filter_fields'):
            raise ImproperlyConfigured(
                "You need to define `nested_filter_fields` in your `{}` view "
                "when using `{}` filter backend."
                "".format(view.__class__.__name__, cls.__name__)
            )

        filter_fields = view.nested_filter_fields

        for field, options in filter_fields.items():
            if options is None or isinstance(options, string_types):
                filter_fields[field] = {
                    'field': options or field
                }
            elif 'field' not in filter_fields[field]:
                filter_fields[field]['field'] = field

            if 'lookups' not in filter_fields[field]:
                filter_fields[field]['lookups'] = tuple(
                    ALL_LOOKUP_FILTERS_AND_QUERIES
                )

        return filter_fields

    def get_filter_field_nested_path(self, filter_fields, field_name):
        """Get filter field path to be used in nested query.

        :param filter_fields:
        :param field_name:
        :return:
        """
        if 'path' in filter_fields[field_name]:
            return filter_fields[field_name]['path']
        return field_name

    def get_filter_query_params(self, request, view):
        """Get query params to be filtered on.

        :param request: Django REST framework request.
        :param view: View.
        :type request: rest_framework.request.Request
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Request query params to filter on.
        :rtype: dict
        """
        query_params = request.query_params.copy()

        filter_query_params = {}
        filter_fields = self.prepare_filter_fields(view)
        for query_param in query_params:
            query_param_list = self.split_lookup_filter(
                query_param,
                maxsplit=1
            )
            field_name = query_param_list[0]

            if field_name in filter_fields:
                lookup_param = None
                if len(query_param_list) > 1:
                    lookup_param = query_param_list[1]

                valid_lookups = filter_fields[field_name]['lookups']
                nested_path = self.get_filter_field_nested_path(
                    filter_fields,
                    field_name
                )

                if lookup_param is None or lookup_param in valid_lookups:
                    values = [
                        __value.strip()
                        for __value
                        in query_params.getlist(query_param)
                        if __value.strip() != ''
                    ]

                    if values:
                        filter_query_params[query_param] = {
                            'lookup': lookup_param,
                            'values': values,
                            'field': filter_fields[field_name].get(
                                'field',
                                field_name
                            ),
                            'type': view.mapping,
                            'path': nested_path,
                        }
        return filter_query_params

    @classmethod
    def apply_filter(cls, queryset, options=None, args=None, kwargs=None):
        """Apply filter.

        :param queryset:
        :param options:
        :param args:
        :param kwargs:
        :return:
        """
        if options is None or 'path' not in options:
            raise ImproperlyConfigured(
                "You should provide an `path` argument in the field options."
            )

        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}

        return queryset.query(
            'nested',
            path=options.get('path'),
            query=Q(*args, **kwargs)
        )

    @classmethod
    def apply_query(cls, queryset, options=None, args=None, kwargs=None):
        """Apply query.

        :param queryset:
        :param options:
        :param args:
        :param kwargs:
        :return:
        """
        if options is None:
            raise ImproperlyConfigured(
                "You should provide an `path` argument in the field options."
            )

        path = options.pop('path')

        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}

        return queryset.query(
            'nested',
            path=path,
            query=Q(*args, **kwargs)
        )

    def get_coreschema_field(self, field):
        if isinstance(field, fields.IntegerField):
            field_cls = coreschema.Number
        else:
            field_cls = coreschema.String
        return field_cls()

    def get_schema_fields(self, view):
        assert coreapi is not None, 'coreapi must be installed to ' \
                                    'use `get_schema_fields()`'
        assert coreschema is not None, 'coreschema must be installed to ' \
                                       'use `get_schema_fields()`'
        filter_fields = getattr(view, 'nested_filter_fields', None)
        document = getattr(view, 'document', None)

        return [] if not filter_fields else [
            coreapi.Field(
                name=field_name,
                required=False,
                location='query',
                schema=self.get_coreschema_field(
                    document._fields.get(field_name)
                )
            )
            for field_name in filter_fields
        ]



"""
Term level filtering and ``post_filter`` backends.
"""

from .common import FilteringFilterBackend
from .geo_spatial import GeoSpatialFilteringFilterBackend
from .ids import IdsFilterBackend
from .nested import NestedFilteringFilterBackend
from .post_filter import PostFilterFilteringFilterBackend

__title__ = 'django_elasticsearch_dsl_drf.filter_backends.filtering'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'FilteringFilterBackend',
    'GeoSpatialFilteringFilterBackend',
    'IdsFilterBackend',
    'NestedFilteringFilterBackend',
    'PostFilterFilteringFilterBackend',
)



"""
Test utils.
"""

from __future__ import absolute_import

import unittest

from django.urls import reverse
from django.core.management import call_command

import pytest

from rest_framework import status

from .base import BaseRestFrameworkTestCase

__title__ = 'django_elasticsearch_dsl_drf.tests.test_utils'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'TestUtils',
)


@pytest.mark.django_db
class TestUtils(BaseRestFrameworkTestCase):
    """Test utils."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        super(TestUtils, cls).setUpClass()

        cls.sleep()
        call_command('search_index', '--rebuild', '-f')

    def _list_results(self):
        """List results."""
        self.authenticate()

        url = reverse('bookdocument_no_records-list', kwargs={})

        # Make request
        response = self.client.get(url, {})
        self.assertEqual(
            response.status_code,
            status.HTTP_200_OK
        )

        # Should contain no results
        self.assertEqual(response.data['results'], [])
        self.assertEqual(response.data['count'], 0)
        self.assertEqual(response.data['next'], None)
        self.assertEqual(response.data['previous'], None)

    def test_list_results(self):
        """Test list results."""
        return self._list_results()


if __name__ == '__main__':
    unittest.main()



import operator
import six

from elasticsearch_dsl.query import Q

from .base import BaseSearchQueryBackend

__title__ = 'django_elasticsearch_dsl_drf.filter_backends.search.' \
            'query_backends.nested'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = ('NestedQueryBackend',)


class NestedQueryBackend(BaseSearchQueryBackend):
    """Nested query backend."""

    query_type = 'nested'

    @classmethod
    def construct_search(cls, request, view, search_backend):
        """Construct search.

        Dictionary key is the GET param name. The path option stands for the
        path in Elasticsearch.

        Type 1:

            search_nested_fields = {
                'country': {
                    'path': 'country',
                    'fields': ['name'],
                },
                'city': {
                    'path': 'country.city',
                    'fields': ['name'],
                },
            }

        Type 2:

            search_nested_fields = {
                'country': {
                    'path': 'country',
                    'fields': [{'name': {'boost': 2}}]
                },
                'city': {
                    'path': 'country.city',
                    'fields': [{'name': {'boost': 2}}]
                },
            }

        :param request:
        :param view:
        :param search_backend:
        :return:
        """
        if not hasattr(view, 'search_nested_fields'):
            return []

        query_params = search_backend.get_search_query_params(request)
        __queries = []

        for search_term in query_params:
            __values = search_backend.split_lookup_name(search_term, 1)
            __len_values = len(__values)
            if __len_values > 1:
                label, value = __values
                if label in view.search_nested_fields:
                    options = view.search_nested_fields.get(label)
                    path = options.get('path')

                    queries = []
                    for _field in options.get('fields', []):
                        # In case if we deal with structure 2
                        if isinstance(_field, dict):
                            # take options (ex: boost) into consideration
                            field_options = {key: value for key, value in _field.items() if key != 'name'}
                            field_options.update({
                                "query": search_term,
                            })
                            field = "{}.{}".format(path, _field['name'])
                            field_kwargs = {
                                field: field_options,
                            }
                        # In case if we deal with structure 1
                        else:
                            field = "{}.{}".format(path, _field)
                            field_kwargs = {
                                field: value
                            }

                        queries = [
                            Q("match", **field_kwargs)
                        ]

                    __queries.append(
                        Q(
                            cls.query_type,
                            path=path,
                            query=six.moves.reduce(operator.or_, queries)
                        )
                    )
            else:
                for label, options in view.search_nested_fields.items():
                    queries = []
                    path = options.get('path')

                    for _field in options.get('fields', []):
                        # In case if we deal with structure 2
                        if isinstance(_field, dict):
                            # take options (ex: boost) into consideration
                            field_options = {key: value for key, value in _field.items() if key != 'name'}
                            field_options.update({
                                "query": search_term,
                            })
                            field = "{}.{}".format(path, _field['name'])
                            field_kwargs = {
                                field: field_options,
                            }
                        # In case if we deal with structure 1
                        else:
                            field = "{}.{}".format(path, _field)
                            field_kwargs = {
                                field: search_term
                            }

                        queries.append(
                            Q("match", **field_kwargs)
                        )

                    __queries.append(
                        Q(
                            cls.query_type,
                            path=path,
                            query=six.moves.reduce(operator.or_, queries)
                        )
                    )

        return __queries



"""
Suggester filtering backends.
"""

from .functional import FunctionalSuggesterFilterBackend
from .native import SuggesterFilterBackend

__title__ = 'django_elasticsearch_dsl_drf.filter_backends.suggester'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'SuggesterFilterBackend',
    'FunctionalSuggesterFilterBackend',
)



# -*- coding: utf-8 -*-
"""
Test pip_helpers.
"""

from __future__ import absolute_import, unicode_literals

import unittest

import django
import pytest

from ..pip_helpers import check_if_installed, get_installed_packages

__title__ = 'django_elasticsearch_dsl_drf.tests.test_pip_helpers'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'TestPipHelpers',
)


@pytest.mark.django_db
class TestPipHelpers(unittest.TestCase):
    """Test pip_helpers."""

    @classmethod
    def setUpClass(cls):
        cls.mapping = {
            'country': {
                'name': 'Netherlands',
                'province': {
                    'name': 'North Holland',
                    'city': {
                        'name': 'Amsterdam',
                    }
                }
            }
        }

    def test_get_installed_packages(self):
        """Test `get_installed_packages`.

        :return:
        """
        installed_packages = get_installed_packages()
        self.assertIn('Django', installed_packages)
        self.assertIn('elasticsearch', installed_packages)
        self.assertIn('elasticsearch-dsl', installed_packages)

    def test_get_installed_packages_with_versions(self):
        """Test `get_installed_packages`.

        :return:
        """
        installed_packages = get_installed_packages(with_versions=True)
        django_version = django.get_version()
        self.assertIn(('Django', django_version), installed_packages)

    def test_check_if_installed(self):
        """Test `check_if_installed`.

        :return:
        """
        self.assertTrue(check_if_installed('Django'))
        self.assertTrue(check_if_installed('elasticsearch'))
        self.assertTrue(check_if_installed('elasticsearch-dsl'))
        self.assertFalse(check_if_installed('django-fobi'))



"""
Search query backends.
"""

from .base import BaseSearchQueryBackend
from .match import MatchQueryBackend
from .match_phrase import MatchPhraseQueryBackend
from .match_phrase_prefix import MatchPhrasePrefixQueryBackend
from .multi_match import MultiMatchQueryBackend
from .nested import NestedQueryBackend
from .simple_query_string import SimpleQueryStringQueryBackend

__title__ = 'django_elasticsearch_dsl_drf.filter_backends.search.' \
            'query_backends'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'BaseSearchQueryBackend',
    'MatchQueryBackend',
    'MatchPhraseQueryBackend',
    'MatchPhrasePrefixQueryBackend',
    'MultiMatchQueryBackend',
    'NestedQueryBackend',
    'SimpleQueryStringQueryBackend',
)



# -*- coding: utf-8 -*-
"""
Test wrappers.
"""

from __future__ import absolute_import, unicode_literals

import json
import unittest

import pytest

from ..wrappers import obj_to_dict, dict_to_obj

__title__ = 'django_elasticsearch_dsl_drf.tests.test_wrappers'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'TestWrappers',
)


@pytest.mark.django_db
class TestWrappers(unittest.TestCase):
    """Test wrappers."""

    @classmethod
    def setUpClass(cls):
        cls.mapping = {
            'country': {
                'name': 'Netherlands',
                'province': {
                    'name': 'North Holland',
                    'city': {
                        'name': 'Amsterdam',
                    }
                }
            }
        }

    def test_obj_to_dict(self):
        """Test `obj_to_dict`.

        :return:
        """
        wrapper = dict_to_obj(self.mapping)
        self.assertEqual(
            wrapper.country.name,
            self.mapping['country']['name']
        )

        self.assertEqual(
            wrapper.country.province.name,
            self.mapping['country']['province']['name']
        )

        self.assertEqual(
            wrapper.country.province.city.name,
            self.mapping['country']['province']['city']['name']
        )

        # See if original ``mapping`` is same as ``as_dict``
        self.assertEqual(
            wrapper.as_dict,
            self.mapping
        )

        # Since we don't know for sure which one will be, we need to make
        # sure it's one of the items.
        self.assertIn(
            str(wrapper),
            (
                self.mapping['country']['name'],
                self.mapping['country']['province']['name'],
                self.mapping['country']['province']['city']['name'],
            )
        )

    def test_dict_to_obj(self):
        """Test `dict_to_obj`.

        :return:
        """
        wrapper = dict_to_obj(self.mapping)
        mapping = obj_to_dict(wrapper)
        self.assertEqual(self.mapping, mapping)

    def test_wrapper_as_json(self):
        """Test :Wrapper:`as_json` property."""
        wrapper = dict_to_obj(self.mapping)
        self.assertEqual(
            json.loads(wrapper.as_json),
            json.loads(
                '{"country": '
                '{"name": "Netherlands", '
                '"province": {'
                '"name": "North Holland", '
                '"city": {'
                '"name": "Amsterdam"}}}}'
            )
        )



"""
Test views.
"""

from __future__ import absolute_import

import unittest

from django.core.management import call_command
from django.urls import reverse

import pytest

from rest_framework import status

import factories

from .base import BaseRestFrameworkTestCase

__title__ = 'django_elasticsearch_dsl_drf.tests.test_views'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'TestViews',
)


@pytest.mark.django_db
class TestViews(BaseRestFrameworkTestCase):
    """Test views."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        """Set up class."""
        super(TestViews, cls).setUpClass()

        cls.books = factories.BookWithoutTagsAndOrdersFactory.create_batch(20)
        cls.tags = factories.TagGenreFactory.create_batch(20)

        cls.sleep()
        call_command('search_index', '--rebuild', '-f')

    def test_listing_view(self):
        """Test listing view."""
        url = reverse('bookdocument-list', kwargs={})
        data = {}

        # Should contain 20 results
        response = self.client.get(url, data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data['results']), 20)

    def test_detail_view(self):
        """Test detail view."""
        __obj = self.books[0]
        url = reverse('bookdocument-detail', kwargs={'id': __obj.pk})
        data = {}

        response = self.client.get(url, data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        for __field in ('id', 'title', 'pages', 'state', 'isbn'):
            self.assertEqual(
                response.data[__field],
                getattr(__obj, __field)
            )

    def test_detail_view_with_custom_lookup(self):
        """Test detail view with a custom lookup field."""
        __obj = self.tags[0]
        url = reverse('tagdocument-detail', kwargs={'title': __obj.title})
        data = {}

        response = self.client.get(url, data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['id'], __obj.title)


if __name__ == '__main__':
    unittest.main()



"""
Search filter backends.
"""

from .base import BaseSearchFilterBackend
from .compound import CompoundSearchFilterBackend
from .historical import SearchFilterBackend
from .multi_match import MultiMatchSearchFilterBackend
from .simple_query_string import SimpleQueryStringSearchFilterBackend

__title__ = 'django_elasticsearch_dsl_drf.filter_backends.search'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'BaseSearchFilterBackend',
    'CompoundSearchFilterBackend',
    'MultiMatchSearchFilterBackend',
    'SearchFilterBackend',
    'SimpleQueryStringSearchFilterBackend',
)



"""
Test pagination.
"""

from __future__ import absolute_import

import unittest

from django.core.management import call_command
from django.urls import reverse

import pytest

from rest_framework import status

import factories

from .base import BaseRestFrameworkTestCase

__title__ = 'django_elasticsearch_dsl_drf.tests.test_pagination'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'TestPagination',
)


@pytest.mark.django_db
class TestPagination(BaseRestFrameworkTestCase):
    """Test pagination."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        """Set up class."""
        super(TestPagination, cls).setUpClass()
        cls.publishers = factories.PublisherFactory.create_batch(40)
        cls.books = factories.BookFactory.create_batch(40)

        cls.sleep()
        call_command('search_index', '--rebuild', '-f')

    def _test_pagination(self):
        """Test pagination."""
        self.authenticate()

        publishers_url = reverse('publisherdocument-list', kwargs={})
        books_url = reverse('bookdocument-list', kwargs={})
        data = {}

        invalid_page_url = books_url + '?page=3&page_size=30'

        invalid_response = self.client.get(invalid_page_url, data)
        self.assertEqual(
            invalid_response.status_code,
            status.HTTP_404_NOT_FOUND
        )

        valid_page_url = publishers_url + '?limit=5&offset=8'

        # Check if response now is valid
        valid_response = self.client.get(valid_page_url, data)
        self.assertEqual(valid_response.status_code, status.HTTP_200_OK)

        # Check totals
        self.assertEqual(len(valid_response.data['results']), 5)

    def test_pagination(self):
        """Test pagination."""
        return self._test_pagination()


if __name__ == '__main__':
    unittest.main()



"""
Nested fields.
"""

from rest_framework.serializers import Field
from .helpers import to_representation

__title__ = 'django_elasticsearch_dsl_drf.fields.nested_fields'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'GeoPointField',
    'GeoShapeField',
    'NestedField',
    'ObjectField',
    'ListField',
)


class ObjectField(Field):
    """Object field."""

    def get_value(self, dictionary):
        """Get value."""
        value = super(ObjectField, self).get_value(dictionary)

        return to_representation(value)

    def to_internal_value(self, data):
        """To internal value."""
        return data

    def to_representation(self, value):
        """To representation."""
        return to_representation(value)


class NestedField(ObjectField):
    """Nested field."""


class GeoPointField(ObjectField):
    """Geo point field."""


class GeoShapeField(ObjectField):
    """Geo shape field."""


class ListField(Field):
    """List field."""

    def get_value(self, dictionary):
        """Get value."""
        value = super(ListField, self).get_value(dictionary)
        return to_representation(value)

    def to_internal_value(self, data):
        """To internal value."""
        return data

    def to_representation(self, value):
        """To representation."""
        return to_representation(value)



"""
Integrate Elasticsearch DSL with Django REST framework.
"""

__title__ = 'django-elasticsearch-dsl-drf'
__version__ = '0.22.5'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2022 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'


from django_nine import versions

if versions.DJANGO_LTE_3_1:
    __all__ = ('default_app_config',)
    default_app_config = 'django_elasticsearch_dsl_drf.apps.Config'



"""
Test filtering `post_filter` backend.
"""

from __future__ import absolute_import

import unittest

from django.core.management import call_command
from django.urls import reverse

import pytest

from rest_framework import status

from books import constants
from search_indexes.viewsets import BookDocumentViewSet

from ..constants import SEPARATOR_LOOKUP_COMPLEX_VALUE
from ..filter_backends import PostFilterFilteringFilterBackend
from .base import (
    BaseRestFrameworkTestCase,
    CORE_API_AND_CORE_SCHEMA_ARE_INSTALLED,
    CORE_API_AND_CORE_SCHEMA_MISSING_MSG,
)
from .data_mixins import AddressesMixin, BooksMixin

__title__ = 'django_elasticsearch_dsl_drf.tests.test_filtering_post'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'TestFilteringPostFilter',
)


@pytest.mark.django_db
class TestFilteringPostFilter(BaseRestFrameworkTestCase,
                              AddressesMixin,
                              BooksMixin):
    """Test filtering `post_filter`."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        """Set up."""
        super(TestFilteringPostFilter, cls).setUpClass()
        # Testing simple documents: Publisher index.
        cls.create_books()

        # Testing nested objects: Addresses, cities and countries
        cls.created_addresses()

        cls.sleep()
        # Update the Elasticsearch index
        call_command('search_index', '--rebuild', '-f')

        # Testing coreapi and coreschema
        cls.backend = PostFilterFilteringFilterBackend()
        cls.view = BookDocumentViewSet()

    # ***********************************************************************
    # ************************ Simple fields ********************************
    # ***********************************************************************

    def _field_filter_value(self, field_name, value, count):
        """Field filter value.

        Usage example:

            >>> self._field_filter_value(
            >>>     'title__wildcard',
            >>>     self.prefix[3:-3],
            >>>     self.prefix_count
            >>> )
        """
        url = self.base_url[:]
        data = {}
        response = self.client.get(
            url + '?{}={}'.format(field_name, value),
            data
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            len(response.data['results']),
            count
        )

    def _field_filter_term(self, field_name, filter_value):
        """Field filter term.

        Example:

            http://localhost:8000/api/articles/?tags=children
        """
        self.authenticate()

        url = self.base_url[:]
        data = {}

        # Should contain 22 results
        response = self.client.get(url, data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data['results']), self.all_count)

        # Should contain only 10 results
        filtered_response = self.client.get(
            url + '?{}={}'.format(field_name, filter_value),
            data
        )
        self.assertEqual(filtered_response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            len(filtered_response.data['results']),
            self.published_count
        )

    def test_field_filter_term(self):
        """Field filter term."""
        return self._field_filter_term(
            'state',
            constants.BOOK_PUBLISHING_STATUS_PUBLISHED
        )

    def test_field_filter_term_explicit(self):
        """Field filter term."""
        return self._field_filter_term(
            'state__term',
            constants.BOOK_PUBLISHING_STATUS_PUBLISHED
        )

    def test_field_filter_range(self):
        """Field filter range.

        Example:

            http://localhost:8000/api/users/?age__range=16;67
        """
        lower_id = self.published[0].id
        upper_id = self.published[-1].id
        value = '{lower_id}{sep}{upper_id}'.format(
            lower_id=lower_id,
            sep=SEPARATOR_LOOKUP_COMPLEX_VALUE,
            upper_id=upper_id,
        )
        return self._field_filter_value(
            'id__range',
            value,
            self.published_count
        )

    def test_field_filter_range_with_boost(self):
        """Field filter range.

        Example:

            http://localhost:8000/api/users/?age__range=16;67;2.0
        """
        lower_id = self.published[0].id
        upper_id = self.published[-1].id
        value = '{lower_id}{sep}{upper_id}{sep}{boost}'.format(
            lower_id=lower_id,
            upper_id=upper_id,
            boost='2.0',
            sep=SEPARATOR_LOOKUP_COMPLEX_VALUE,
        )
        return self._field_filter_value(
            'id__range',
            value,
            self.published_count
        )

    def test_field_filter_prefix(self):
        """Test filter prefix.

        Example:

            http://localhost:8000/api/articles/?tags__prefix=bio
        """
        return self._field_filter_value(
            'title__prefix',
            self.prefix,
            self.prefix_count
        )

    def test_field_filter_in(self):
        """Test filter in.

        Example:

            http://localhost:8000/api/articles/?id__in=1;2;3
        """
        return self._field_filter_value(
            'id__in',
            SEPARATOR_LOOKUP_COMPLEX_VALUE.join(
                [str(__b.id) for __b in self.prefixed]
            ),
            self.prefix_count
        )

    def _field_filter_terms_list(self, field_name, in_values, count):
        """Field filter terms.

        Example:

            http://localhost:8000/api/articles/?id=1&id=2&id=3
        """
        url = self.base_url[:]
        data = {}
        url_parts = ['{}={}'.format(field_name, val) for val in in_values]
        response = self.client.get(
            url + '?{}'.format('&'.join(url_parts)),
            data
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            len(response.data['results']),
            count
        )

    def test_field_filter_terms_list(self):
        """Test filter terms."""
        return self._field_filter_terms_list(
            'id',
            [str(__b.id) for __b in self.prefixed],
            self.prefix_count
        )

    def test_field_filter_terms_string(self):
        """Test filter terms.

        Example:

            http://localhost:8000/api/articles/?id__terms=1;2;3
        """
        return self._field_filter_value(
            'id__terms',
            SEPARATOR_LOOKUP_COMPLEX_VALUE.join(
                [str(__b.id) for __b in self.prefixed]
            ),
            self.prefix_count
        )

    def test_field_filter_exists_true(self):
        """Test filter exists true.

        Example:

            http://localhost:8000/api/articles/?tags__exists=true
        """
        return self._field_filter_value(
            'tags__exists',
            'true',
            self.all_count
        )

    def test_field_filter_exists_false(self):
        """Test filter exists.

        Example:

            http://localhost:8000/api/articles/?non_existent__exists=false
        """
        return self._field_filter_value(
            'non_existent_field__exists',
            'false',
            self.all_count
        )

    def test_field_filter_wildcard(self):
        """Test filter wildcard.

        Example:

            http://localhost:8000/api/articles/?title__wildcard=*elusional*
        """
        return self._field_filter_value(
            'title__wildcard',
            '*{}*'.format(self.prefix[1:6]),
            self.prefix_count
        )

    def test_field_filter_exclude(self):
        """Test filter exclude.

        Example:

            http://localhost:8000/api/articles/?tags__exclude=children
        """
        return self._field_filter_value(
            'state__exclude',
            constants.BOOK_PUBLISHING_STATUS_PUBLISHED,
            self.all_count - self.published_count
        )

    def test_field_filter_isnull_true(self):
        """Test filter isnull true.

        Example:

            http://localhost:8000/api/articles/?null_field__isnull=true
        """
        self._field_filter_value(
            'null_field__isnull',
            'true',
            self.all_count
        )
        self._field_filter_value(
            'tags__isnull',
            'true',
            self.no_tags_count
        )

    def test_field_filter_isnull_false(self):
        """Test filter isnull true.

        Example:

            http://localhost:8000/api/articles/?tags__isnull=false
        """
        self._field_filter_value(
            'null_field__isnull',
            'false',
            0
        )
        self._field_filter_value(
            'tags__isnull',
            'false',
            self.all_count - self.no_tags_count
        )

    def test_field_filter_endswith(self):
        """Test filter endswith.

        Example:

            http://localhost:8000/api/articles/?state__endswith=lished
        """
        return self._field_filter_value(
            'state__endswith',
            constants.BOOK_PUBLISHING_STATUS_PUBLISHED[4:],
            self.published_count
        )

    def test_field_filter_contains(self):
        """Test filter contains.

        Example:

            http://localhost:8000/api/articles/?state__contains=lishe
        """
        return self._field_filter_value(
            'state__contains',
            constants.BOOK_PUBLISHING_STATUS_PUBLISHED[4:-2],
            self.published_count
        )

    def _field_filter_gte_lte(self, field_name, value, lookup, boost=None):
        """Field filter gt/gte/lt/lte.

        Example:

            http://localhost:8000/api/users/?id__gt=10
            http://localhost:8000/api/users/?id__gte=10
            http://localhost:8000/api/users/?id__lt=10
            http://localhost:8000/api/users/?id__lte=10
        """
        url = self.base_url[:]
        data = {}

        if boost is not None:
            url += '?{field_name}__{lookup}={value}{sep}{boost}'.format(
                field_name=field_name,
                lookup=lookup,
                value=value,
                boost=boost,
                sep=SEPARATOR_LOOKUP_COMPLEX_VALUE,
            )
        else:
            url += '?{}__{}={}'.format(field_name, lookup, value)

        response = self.client.get(
            url,
            data
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        __mapping = {
            'gt': self.assertGreater,
            'gte': self.assertGreaterEqual,
            'lt': self.assertLess,
            'lte': self.assertLessEqual,
        }
        __func = __mapping.get(lookup)

        if callable(__func):
            for obj in response.data['results']:
                __func(
                    obj['id'],
                    value
                )

    def test_field_filter_gt(self):
        """Field filter gt.

        Example:

            http://localhost:8000/api/users/?id__gt=10
        :return:
        """
        return self._field_filter_gte_lte('id', self.in_progress[0].id, 'gt')

    def test_field_filter_gt_with_boost(self):
        """Field filter gt with boost.

        Example:

            http://localhost:8000/api/users/?id__gt=10;2.0
        :return:
        """
        # TODO: check boost value
        return self._field_filter_gte_lte(
            'id',
            self.in_progress[0].id,
            'gt',
            '2.0'
        )

    def test_field_filter_gte(self):
        """Field filter gte.

        Example:

            http://localhost:8000/api/users/?id__gte=10
        :return:
        """
        return self._field_filter_gte_lte('id', self.in_progress[0].id, 'gte')

    def test_field_filter_lt(self):
        """Field filter lt.

        Example:

            http://localhost:8000/api/users/?id__lt=10
        :return:
        """
        return self._field_filter_gte_lte('id', self.in_progress[0].id, 'lt')

    def test_field_filter_lt_with_boost(self):
        """Field filter lt with boost.

        Example:

            http://localhost:8000/api/users/?id__lt=10;2.0
        :return:
        """
        # TODO: check boost value
        return self._field_filter_gte_lte(
            'id',
            self.in_progress[0].id,
            'lt',
            '2.0'
        )

    def test_field_filter_lte(self):
        """Field filter lte.

        Example:

            http://localhost:8000/api/users/?id__lte=10
        :return:
        """
        return self._field_filter_gte_lte('id', self.in_progress[0].id, 'lte')

    def test_ids_filter(self):
        """Test ids filter.

        Example:

            http://localhost:8000/api/articles/?ids=68;64;58
            http://localhost:8000/api/articles/?ids=68&ids=64&ids=58
        """
        __ids = [str(__obj.id) for __obj in self.published]
        return self._field_filter_value(
            'ids',
            SEPARATOR_LOOKUP_COMPLEX_VALUE.join(__ids),
            self.published_count
        )

    # ***********************************************************************
    # ************************** Test facets ********************************
    # ***********************************************************************

    # This is what's it's all about - the facets.
    def _list_results_with_facets(self):
        """List results with facets."""
        self.authenticate()

        url = reverse('bookdocument-list', kwargs={})
        facet_state_url = url + '?facet=state'
        data = {}

        # ******************************************************************
        # ************************ No args response ************************
        # ******************************************************************

        # Make request
        no_args_response = self.client.get(url, data)
        self.assertEqual(no_args_response.status_code, status.HTTP_200_OK)

        # Should contain `self.all_count` results
        self.assertEqual(len(no_args_response.data['results']), self.all_count)

        # Should contain 1 facets
        self.assertEqual(len(no_args_response.data['facets']), 1)

        # ******************************************************************
        # ********************** With facets response **********************
        # ******************************************************************

        # Make request
        facet_state_response = self.client.get(facet_state_url, data)
        self.assertEqual(facet_state_response.status_code, status.HTTP_200_OK)

        # Should contain `self.all_count` results
        self.assertEqual(
            len(facet_state_response.data['results']), self.all_count
        )

        # Should contain 2 facets
        self.assertEqual(len(facet_state_response.data['facets']), 2)
        # With 3 statuses
        self.assertEqual(
            len(facet_state_response.data['facets']
                ['_filter_state']['state']['buckets']),
            3
        )

        self.assertIn('_filter_publisher', facet_state_response.data['facets'])
        self.assertIn(
            'publisher',
            facet_state_response.data['facets']['_filter_publisher']
        )

        self.assertIn('_filter_state', facet_state_response.data['facets'])
        self.assertIn(
            'state',
            facet_state_response.data['facets']['_filter_state']
        )
        self.assertIn(
            'buckets',
            facet_state_response.data['facets']['_filter_state']['state']
        )
        # self.assertIn(
        #     'buckets',
        #     facet_state_response.data['facets']['_filter_state']['state']
        # )
        self.assertIn(
            {
                "doc_count": 10,
                "key": "published"
            },
            facet_state_response.data['facets']
                                     ['_filter_state']
                                     ['state']
                                     ['buckets']
        )
        self.assertIn(
            {
                "doc_count": 10,
                "key": "in_progress"
            },
            facet_state_response.data['facets']
            ['_filter_state']
            ['state']
            ['buckets']
        )
        self.assertIn(
            {
                "doc_count": 7,
                "key": "rejected"
            },
            facet_state_response.data['facets']
            ['_filter_state']
            ['state']
            ['buckets']
        )

        # ******************************************************************
        # ******************* With facets filtered response ****************
        # ******************************************************************

        facet_state_filtered_url = url + '?facet=state&state_pf=published'
        # Make request
        facet_state_filtered_response = self.client.get(
            facet_state_filtered_url,
            data
        )
        self.assertEqual(
            facet_state_filtered_response.status_code,
            status.HTTP_200_OK
        )

        # Should contain `self.published_count` results
        self.assertEqual(
            len(facet_state_filtered_response.data['results']),
            self.published_count
        )

        # Should contain 2 facets
        self.assertEqual(len(facet_state_filtered_response.data['facets']), 2)
        # With 3 statuses
        self.assertEqual(
            len(facet_state_filtered_response.data['facets']
                ['_filter_state']['state']['buckets']),
            3
        )

        # Still same facets
        self.assertIn(
            {
                "doc_count": self.published_count,
                "key": "published"
            },
            facet_state_response.data['facets']
            ['_filter_state']
            ['state']
            ['buckets']
        )
        self.assertIn(
            {
                "doc_count": self.in_progress_count,
                "key": "in_progress"
            },
            facet_state_response.data['facets']
            ['_filter_state']
            ['state']
            ['buckets']
        )
        self.assertIn(
            {
                "doc_count": self.rejected_count,
                "key": "rejected"
            },
            facet_state_response.data['facets']
            ['_filter_state']
            ['state']
            ['buckets']
        )

    def test_list_results_with_facets(self):
        """Test list results with facets."""
        return self._list_results_with_facets()

    # ***********************************************************************
    # ************************* Other fields ********************************
    # ***********************************************************************

    def test_various_complex_fields(self):
        """Test various complex fields.

        :return:
        """
        data = {}
        response = self.client.get(self.cities_url, data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        response = self.client.get(self.city_detail_url, data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    # ***********************************************************************
    # ******************** Core api and core schema *************************
    # ***********************************************************************

    @unittest.skipIf(not CORE_API_AND_CORE_SCHEMA_ARE_INSTALLED,
                     CORE_API_AND_CORE_SCHEMA_MISSING_MSG)
    def test_schema_fields_with_filter_fields_list(self):
        """Test schema field generator"""
        fields = self.backend.get_schema_fields(self.view)
        fields = [f.name for f in fields]
        self.assertEqual(fields, list(self.view.post_filter_fields.keys()))

    @unittest.skipIf(not CORE_API_AND_CORE_SCHEMA_ARE_INSTALLED,
                     CORE_API_AND_CORE_SCHEMA_MISSING_MSG)
    def test_schema_field_not_required(self):
        """Test schema fields always not required"""
        fields = self.backend.get_schema_fields(self.view)
        fields = [f.required for f in fields]
        for field in fields:
            self.assertFalse(field)


if __name__ == '__main__':
    unittest.main()



# coding: utf-8
"""
Test more-like-this functionality.
"""

from __future__ import absolute_import, unicode_literals

import unittest
import logging

from django.core.management import call_command
from django.urls import reverse

import pytest

from rest_framework import status

import factories

from .base import BaseRestFrameworkTestCase

__title__ = 'django_elasticsearch_dsl_drf.tests.test_suggesters'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'TestMoreLikeThis',
)

LOGGER = logging.getLogger(__name__)


@pytest.mark.django_db
class TestMoreLikeThis(BaseRestFrameworkTestCase):
    """Test suggesters."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        """Set up class."""
        super(TestMoreLikeThis, cls).setUpClass()

        cls.lorem_books = factories.BookFactory.create_batch(200)

        cls.books = []

        cls.alice_books = []
        for book_data in factories.constants.NON_FAKER_BOOK_CONTENT:
            cls.alice_books.append(
                factories.BookChapterFactory(
                    title=book_data['title'],
                    summary=book_data['summary'],
                    description=book_data['description'],
                )
            )
        cls.alice_books_ids = [__obj.id for __obj in cls.alice_books]

        cls.sheckley_books = []
        for book_data in factories.constants.NON_FAKER_BOOK_CONTENT_OTHER:
            cls.sheckley_books.append(
                factories.BookNovelFactory(
                    title=book_data['title'],
                    summary=book_data['summary'],
                    description=book_data['description'],
                )
            )
        cls.sheckley_books_ids = [__obj.id for __obj in cls.sheckley_books]

        cls.books = []
        cls.books.extend(cls.alice_books)
        cls.books.extend(cls.sheckley_books)

        # Alice book
        cls.books_url_1 = reverse(
            'bookdocument_more_like_this-more-like-this',
            kwargs={'id': cls.books[0].id}
        )
        # Sheckley book
        cls.books_url_2 = reverse(
            'bookdocument_more_like_this-more-like-this',
            kwargs={'id': cls.books[-1].id}
        )

        # Alice book
        cls.books_url_1_no_options = reverse(
            'bookdocument_more_like_this_no_options-more-like-this',
            kwargs={'id': cls.books[0].id}
        )
        # Sheckley book
        cls.books_url_2_no_options = reverse(
            'bookdocument_more_like_this_no_options-more-like-this',
            kwargs={'id': cls.books[-1].id}
        )

        cls.sleep()
        call_command('search_index', '--rebuild', '-f')

    def _test_more_like_this(self, test_data_ids, url, strict=True):
        """Test more-like-this.

        We can't really predict which result would it show as most relevant,
        however we can for sure assume that no strange data shall appear
        in between.
        """
        self.authenticate()
        data = {}
        response = self.client.get(
            url,
            data
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        if strict:
            response_ids = [int(__r['id']) for __r in response.data['results']]

            self.assertFalse(
                bool(
                    set(response_ids) - set(test_data_ids)
                )
            )

    def test_more_like_this(self):
        """Test more-like-this."""
        # Testing publishers
        test_data_ids = []
        test_data_ids.extend(self.alice_books_ids)
        test_data_ids.extend(self.sheckley_books_ids)

        self._test_more_like_this(
            test_data_ids,
            self.books_url_1,
            strict=True
        )
        self._test_more_like_this(
            test_data_ids,
            self.books_url_1_no_options,
            strict=False
        )

        self._test_more_like_this(
            test_data_ids,
            self.books_url_2,
            strict=True
        )
        self._test_more_like_this(
            test_data_ids,
            self.books_url_2_no_options,
            strict=False
        )


if __name__ == '__main__':
    unittest.main()



"""
Helpers.
"""
from elasticsearch_dsl.utils import AttrDict, AttrList

__title__ = 'django_elasticsearch_dsl_drf.fields.helpers'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = ('to_representation',)


def to_representation(value):
    """To representation."""
    if isinstance(value, AttrDict):
        return value.to_dict()
    if isinstance(value, AttrList):
        _value = [to_representation(__v) for __v in value]
        return _value
        # If approach above doesn't work, replace it with the code below
        # try:
        #     _value = list(value)
        #     json.dumps(_value)
        #     return _value
        # except TypeError:
        #     _value = [to_representation(__v) for __v in value]
        #     return _value
    return value



"""
Fields.
"""

from .common import (
    BooleanField,
    CharField,
    DateField,
    FloatField,
    IntegerField,
    IPAddressField,
)
from .nested_fields import (
    GeoPointField,
    GeoShapeField,
    ListField,
    NestedField,
    ObjectField,
)

__title__ = 'django_elasticsearch_dsl_drf.fields'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'BooleanField',
    'CharField',
    'DateField',
    'FloatField',
    'GeoPointField',
    'GeoShapeField',
    'IntegerField',
    'IPAddressField',
    'ListField',
    'NestedField',
    'ObjectField',
)



"""Base search backend."""

from django.core.exceptions import ImproperlyConfigured

from django_elasticsearch_dsl import fields
from rest_framework.filters import BaseFilterBackend
from rest_framework.settings import api_settings

from ..mixins import FilterBackendMixin
from ...compat import coreapi, coreschema
from ...constants import MATCHING_OPTIONS, DEFAULT_MATCHING_OPTION

__title__ = 'django_elasticsearch_dsl_drf.filter_backends.search.common'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'BaseSearchFilterBackend',
)


class BaseSearchFilterBackend(BaseFilterBackend, FilterBackendMixin):
    """Base search filter backend."""

    query_backends = []

    matching = DEFAULT_MATCHING_OPTION

    search_param = api_settings.SEARCH_PARAM

    def get_search_query_params(self, request):
        """Get search query params.

        :param request: Django REST framework request.
        :type request: rest_framework.request.Request
        :return: List of search query params.
        :rtype: list
        """
        query_params = request.query_params.copy()
        return query_params.getlist(self.search_param, [])

    def get_query_backends(self, request, view):
        """Get query backends.

        :return:
        """
        raise NotImplementedError(
            "You should define `get_query_backends` method in your {} class"
            "".format(self.__class__.__name__)
        )

    def _get_query_backends(self, request, view):
        """Get query backends internal.

        :param request:
        :param view:
        :return:
        """
        try:
            return self.get_query_backends(request, view)
        except NotImplementedError as err:
            pass

        if not self.query_backends:
            raise NotImplementedError(
                "Your search backend shall either implement "
                "`get_query_backends` method or define `query_backends`"
                "property."
            )
        return self.query_backends[:]

    def filter_queryset(self, request, queryset, view):
        """Filter the queryset.

        :param request: Django REST framework request.
        :param queryset: Base queryset.
        :param view: View.
        :type request: rest_framework.request.Request
        :type queryset: elasticsearch_dsl.search.Search
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Updated queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        if self.matching not in MATCHING_OPTIONS:
            raise ImproperlyConfigured(
                "Your `matching` value does not match the allowed matching"
                "options: {}".format(', '.join(MATCHING_OPTIONS))
            )

        __query_backends = self._get_query_backends(request, view)

        if len(__query_backends) > 1:
            __queries = []
            for query_backend in __query_backends:
                __queries.extend(
                    query_backend.construct_search(
                        request=request,
                        view=view,
                        search_backend=self
                    )
                )

            if __queries:
                queryset = queryset.query(
                    'bool',
                    **{self.matching: __queries}
                )

        elif len(__query_backends) == 1:
            __query = __query_backends[0].construct_search(
                request=request,
                view=view,
                search_backend=self
            )
            queryset = queryset.query('bool', **{self.matching: __query})

        else:
            raise ImproperlyConfigured(
                "Search filter backend shall have at least one query_backend"
                "specified either in `query_backends` property or "
                "`get_query_backends` method. Make appropriate changes to"
                "your {} class".format(self.__class__.__name__)
            )

        return queryset

    def get_coreschema_field(self, field):
        if isinstance(field, fields.IntegerField):
            field_cls = coreschema.Number
        else:
            field_cls = coreschema.String
        return field_cls()

    def get_schema_fields(self, view):
        assert coreapi is not None, 'coreapi must be installed to ' \
                                    'use `get_schema_fields()`'
        assert coreschema is not None, 'coreschema must be installed to ' \
                                       'use `get_schema_fields()`'

        _search_fields = getattr(view, 'search_fields', None)
        if isinstance(_search_fields, dict):
            search_fields = list(_search_fields.keys())
        else:
            search_fields = _search_fields

        return [] if not search_fields else [
            coreapi.Field(
                name=self.search_param,
                required=False,
                location='query',
                schema=coreschema.String(
                    description='Search in '
                                '{}.'.format(', '.join(search_fields))
                )
            )
        ]



# -*- coding: utf-8 -*-
"""
Test helpers.
"""

from __future__ import absolute_import, unicode_literals

import unittest

from django.core.management import call_command

import pytest

from ..elasticsearch_helpers import delete_all_indices, get_all_indices

from .base import BaseTestCase

__title__ = 'django_elasticsearch_dsl_drf.tests.test_management_commands'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'TestManagementCommands',
)


@pytest.mark.django_db
class TestManagementCommands(BaseTestCase):
    """Test management commands."""

    pytestmark = pytest.mark.django_db

    def _get_all_indices(self):
        """Get all indices."""
        delete_all_indices(with_protected=True)

        self.sleep()
        call_command('search_index', '--rebuild', '-f')

        res = set(get_all_indices())
        expected = {
            'test_address',
            'test_location',
            'test_publisher',
            'test_journal',
            'test_city',
            'test_author',
            'test_book',
            'test_tag',
        }

        self.assertSetEqual(res, expected)

    def _delete_all_indices(self):
        """Delete all indices."""
        call_command('elasticsearch_remove_indexes', '--with-protected')

        res = get_all_indices()

        self.assertEqual(res, [])

    def test_all(self):
        """Filter by field."""
        self._get_all_indices()
        self._delete_all_indices()


if __name__ == '__main__':
    unittest.main()



"""Multi match search filter backend."""
import logging

from ...constants import MATCHING_OPTION_MUST
from .base import BaseSearchFilterBackend
from .query_backends import (
    MultiMatchQueryBackend,
)

LOGGER = logging.getLogger(__name__)


__title__ = 'django_elasticsearch_dsl_drf.filter_backends.search.multi_match'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'MultiMatchSearchFilterBackend',
)


class MultiMatchSearchFilterBackend(BaseSearchFilterBackend):
    """Multi match search filter backend."""

    search_param = 'search_multi_match'

    matching = MATCHING_OPTION_MUST

    query_backends = [
        MultiMatchQueryBackend,
    ]



"""
Analyzers.
"""
from elasticsearch_dsl import analyzer
from elasticsearch_dsl.analysis import token_filter

__title__ = 'django_elasticsearch_dsl_drf.analyzers'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'edge_ngram_completion_filter',
    'edge_ngram_completion',
)

edge_ngram_completion_filter = token_filter(
    'edge_ngram_completion_filter',
    type="edge_ngram",
    min_gram=1,
    max_gram=20
)


edge_ngram_completion = analyzer(
    "edge_ngram_completion",
    tokenizer="standard",
    filter=["lowercase", edge_ngram_completion_filter]
)



"""
Ordering backends.
"""

from .common import DefaultOrderingFilterBackend, OrderingFilterBackend
from .geo_spatial import GeoSpatialOrderingFilterBackend

__title__ = 'django_elasticsearch_dsl_drf.filter_backends.ordering'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'DefaultOrderingFilterBackend',
    'GeoSpatialOrderingFilterBackend',
    'OrderingFilterBackend',
)



"""
Test multi match search filter backend.
"""

from __future__ import absolute_import

import unittest

from django.core.management import call_command
from django.urls import reverse

import pytest

from rest_framework import status

from books import constants
import factories
from search_indexes.viewsets import (
    BookSimpleQueryStringSearchFilterBackendDocumentViewSet
)
from ..filter_backends import SimpleQueryStringSearchFilterBackend
from ..versions import ELASTICSEARCH_GTE_6_0

from .base import (
    BaseRestFrameworkTestCase,
    CORE_API_AND_CORE_SCHEMA_ARE_INSTALLED,
    CORE_API_AND_CORE_SCHEMA_MISSING_MSG,
)

__title__ = 'django_elasticsearch_dsl_drf.tests.test_search_multi_match'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'TestSimpleQueryStringSearch',
)


@pytest.mark.django_db
class TestSimpleQueryStringSearch(BaseRestFrameworkTestCase):
    """Test simple query string search."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        super(TestSimpleQueryStringSearch, cls).setUpClass()

        # Book factories with unique title
        cls.special_count = 10
        cls.special = factories.BookWithUniqueTitleFactory.create_batch(
            cls.special_count,
            **{
                'summary': 'Delusional Insanity, fine art photography',
                'state': constants.BOOK_PUBLISHING_STATUS_PUBLISHED,
            }
        )

        # Lorem ipsum book factories
        cls.lorem_count = 10
        cls.lorem = factories.BookWithUniqueTitleFactory.create_batch(
            cls.lorem_count
        )

        # Book factories with title, description and summary that actually
        # make sense
        cls.non_lorem_count = 9
        cls.non_lorem = [
            factories.BookChapter20Factory(),
            factories.BookChapter21Factory(),
            factories.BookChapter22Factory(),
            factories.BookChapter60Factory(),
            factories.BookChapter61Factory(),
            factories.BookChapter62Factory(),
            factories.BookChapter110Factory(),
            factories.BookChapter111Factory(),
            factories.BookChapter112Factory(),
        ]

        cls.all_count = (
            cls.special_count + cls.lorem_count + cls.non_lorem_count
        )

        cls.cities_count = 20
        cls.cities = factories.CityFactory.create_batch(cls.cities_count)

        # Create 10 cities in a given country. The reason that we don't
        # do create_batch here is that sometimes in the same test city name is
        # generated twice and thus our concept of precise number matching
        # fails. Before there's a better approach, this would stay so. The
        # create_batch part (below) will remain commented out, until there's a
        # better solution.
        cls.switzerland = factories.CountryFactory.create(name='Wonderland')
        cls.switz_cities_count = 10
        cls.switz_cities_names = [
            'Zurich',
            'Geneva',
            'Basel',
            'Lausanne',
            'Bern',
            'Winterthur',
            'Lucerne',
            'St. Gallen',
            'Lugano',
            'Biel/Bienne',
        ]
        for switz_city in cls.switz_cities_names:
            cls.switz_cities = factories.CityFactory(
                name=switz_city,
                country=cls.switzerland
            )
        # cls.switz_cities = factories.CityFactory.create_batch(
        #     cls.switz_cities_count,
        #     country=cls.switzerland
        # )
        cls.all_cities_count = cls.cities_count + cls.switz_cities_count

        cls.sleep(4)
        call_command('search_index', '--rebuild', '-f')

        # Testing coreapi and coreschema
        cls.backend = SimpleQueryStringSearchFilterBackend()
        cls.view = BookSimpleQueryStringSearchFilterBackendDocumentViewSet()

    def _search(self, search_term, num_results, url=None):
        """Search."""
        self.authenticate()

        if url is None:
            url = reverse(
                'bookdocument_simple_query_string_search_backend-list',
                kwargs={}
            )

        data = {}

        # Should contain 20 results
        response = self.client.get(url, data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data['results']), self.all_count)

        # Should contain only 10 results
        filtered_response = self.client.get(
            url + '?search_simple_query_string={}'.format(search_term),
            data
        )
        self.assertEqual(filtered_response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            len(filtered_response.data['results']),
            num_results
        )

    def _search_boost(self, search_term, ordering, url=None):
        """Search boost.

        In our book view, we have the following defined:

        >>> search_fields = {
        >>>     'title': {'boost': 4},
        >>>     'description': {'boost': 2},
        >>>     'summary': None,
        >>> }

        That means that `title` is more important than `description` and
        `description` is more important than `summary`.
        Results with search term in `title`, `summary` and
        `description` shall be ranked better than results with search term
        in `summary` and `description`. In their turn, results with search term
        in `summary` and `description` shall be ranked better than results
        with search term in `description` only.

        :param search_term:
        :param ordering:
        :return:
        """
        self.authenticate()

        if url is None:
            url = reverse(
                'bookdocument_simple_query_string_search_backend-list',
                kwargs={}
            )
        data = {}

        filtered_response = self.client.get(
            url + '?search_simple_query_string={}'.format(search_term),
            data
        )
        self.assertEqual(filtered_response.status_code, status.HTTP_200_OK)
        self.assertIn('results', filtered_response.data)
        for counter, item_id in enumerate(ordering):
            result_item = filtered_response.data['results'][counter]
            self.assertEqual(result_item['id'], item_id)

    def test_search_with_quotes(self, url=None):
        """Search with quotes."""
        # Search for: "Pig and Pepper"
        self._search(
            search_term='"Pig and Pepper"',
            num_results=1,
            url=url
        )

        # Search for: "Pool of Tears"
        self._search(
            search_term='"Pool of Tears"',
            num_results=3,
            url=url
        )

        # Search for: "Pool of Tears" -considering
        self._search(
            search_term='"Pool of Tears" -considering',
            num_results=1,
            url=url
        )

        # Search for: "chapter II" +fender. Note, that `%2B` stands for `+`
        # symbol in the URL.
        self._search(
            search_term='"chapter II" %2Bfender',
            num_results=1,
            url=url
        )

        # Search for: "chapter II" +shutting. Note, that `%2B` stands for `+`
        # symbol in the URL.
        self._search(
            search_term='"chapter II" %2Bshutting',
            num_results=1,
            url=url
        )

        # Search for: "chapter II" +(shutting | fender). Note, that `%2B`
        # stands for `+` symbol in the URL.
        self._search(
            search_term='"chapter II" %2B(shutting | fender)',
            num_results=2,
            url=url
        )

    @unittest.skipIf(condition=ELASTICSEARCH_GTE_6_0, reason="")
    def test_search_without_quotes(self, url=None):
        """Test search without quotes. This does not work on Elasticsearch 6.x.

        :param url:
        :return:
        """
        # Search for: Pig and Pepper
        self._search(
            search_term='Pig and Pepper',
            num_results=2,
            url=url
        )

    def test_search_with_quotes_boost(self, url=None):
        """Search boost.

        :return:
        """
        # Search for: "Pig and Pepper"
        self._search_boost(
            search_term='"Pig and Pepper"',
            ordering=[
                self.non_lorem[3].pk,
            ],
            url=url
        )

        # Search for: "Pool of Tears"
        self._search_boost(
            search_term='"Pool of Tears"',
            ordering=[
                self.non_lorem[0].pk,
                self.non_lorem[1].pk,
                self.non_lorem[2].pk,
            ],
            url=url
        )

        # Search for: "Pool of Tears" -considering
        self._search_boost(
            search_term='"Pool of Tears" -considering',
            ordering=[
                self.non_lorem[0].pk,
            ],
            url=url
        )

        # Search for: "chapter II" +fender. Note, that `%2B` stands for `+`
        self._search_boost(
            search_term='"chapter II" %2Bfender',
            ordering=[
                self.non_lorem[2].pk,
            ],
            url=url
        )

        # Search for: "chapter II" +shutting. Note, that `%2B` stands for `+`
        # symbol in the URL.
        self._search_boost(
            search_term='"chapter II" %2Bshutting',
            ordering=[
                self.non_lorem[1].pk,
            ],
            url=url
        )

        # Search for: "chapter II" +(shutting | fender). Note, that `%2B`
        # stands for `+` symbol in the URL.
        self._search_boost(
            search_term='"chapter II" %2B(shutting | fender)',
            ordering=[
                self.non_lorem[2].pk,
                self.non_lorem[1].pk,
            ],
            url=url
        )

    @unittest.skipIf(condition=ELASTICSEARCH_GTE_6_0, reason="")
    def test_search_without_quotes_boost(self, url=None):
        """Search boost without quotes. Does not work on Elasticsearch 6.x.

        :return:
        """
        # Search for: Pig and Pepper
        self._search_boost(
            search_term='Pig and Pepper',
            ordering=[
                self.non_lorem[3].pk,
                self.non_lorem[4].pk,
            ],
            url=url
        )

    def test_search_with_quotes_alternative(self):
        """Test search by field.

        :param url:
        :return:
        """
        url = reverse(
            'bookdocument_simple_query_string_boost_search_backend-list',
            kwargs={}
        )
        return self.test_search_with_quotes(url)

    def test_search_with_quotes_boost_alternative(self):
        """Search boost.

        :return:
        """
        url = reverse(
            'bookdocument_simple_query_string_boost_search_backend-list',
            kwargs={}
        )
        return self.test_search_with_quotes_boost(url)

    def test_search_selected_fields(self, url=None):
        """Search boost.

        :return:
        """
        # Search for: "chapter II" +fender. Note, that `%2B` stands for `+`
        # symbol in the URL.
        self._search(
            search_term='title,summary:"Pool of Tears"',
            num_results=2,
            url=url
        )

    def test_search_boost_selected_fields(self, url=None):
        """Search boost.

        :return:
        """
        # Search for (only in `title` and `summary`): "Pool of Tears"
        self._search_boost(
            search_term='title,summary:"Pool of Tears"',
            ordering=[
                self.non_lorem[0].pk,
                self.non_lorem[1].pk,
            ],
            url=url
        )

    @unittest.skipIf(not CORE_API_AND_CORE_SCHEMA_ARE_INSTALLED,
                     CORE_API_AND_CORE_SCHEMA_MISSING_MSG)
    def test_schema_fields_with_filter_fields_list(self):
        """Test schema field generator"""
        fields = self.backend.get_schema_fields(self.view)
        fields = [f.name for f in fields]
        self.assertEqual(fields, ['search'])

    @unittest.skipIf(not CORE_API_AND_CORE_SCHEMA_ARE_INSTALLED,
                     CORE_API_AND_CORE_SCHEMA_MISSING_MSG)
    def test_schema_field_not_required(self):
        """Test schema fields always not required"""
        fields = self.backend.get_schema_fields(self.view)
        fields = [f.required for f in fields]
        for field in fields:
            self.assertFalse(field)


if __name__ == '__main__':
    unittest.main()



"""
Test geo-spatial ordering filter backend.
"""

from __future__ import absolute_import

import unittest

from django.core.management import call_command
from django.urls import reverse

import pytest

from rest_framework import status

import factories

from ..constants import (
    GEO_DISTANCE_ORDERING_PARAM,
    SEPARATOR_LOOKUP_COMPLEX_VALUE,
)
from .base import BaseRestFrameworkTestCase

__title__ = 'django_elasticsearch_dsl_drf.tests.test_filtering'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'TestOrderingGeoSpatial',
)


@pytest.mark.django_db
class TestOrderingGeoSpatial(BaseRestFrameworkTestCase):
    """Test ordering geo-spatial."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        """Set up."""
        super(TestOrderingGeoSpatial, cls).setUpClass()

        cls.geo_origin = factories.PublisherFactory.create(
            **{
                'latitude': 48.8549,
                'longitude': 2.3000,
            }
        )

        cls.geo_in_count = 5
        cls.unit = 'km'
        cls.algo = 'plane'
        cls.geo_in = []
        for index in range(cls.geo_in_count):
            __publisher = factories.PublisherFactory.create(
                **{
                    'latitude': 48.8570 + index,
                    'longitude': 2.3005,
                }
            )
            cls.geo_in.append(__publisher)

        cls.base_publisher_url = reverse('publisherdocument-list', kwargs={})

        cls.sleep()
        call_command('search_index', '--rebuild', '-f')

    @pytest.mark.webtest
    def test_field_filter_geo_distance(self):
        """Field filter geo_distance.

        Example:

            http://localhost:8000
            /api/publisher/?ordering=location;48.85;2.30;km;plane
        """
        self.authenticate()

        __params = 'location{sep}{lat}{sep}{lon}{sep}{unit}{sep}{algo}'.format(
            lat=self.geo_origin.latitude,
            lon=self.geo_origin.longitude,
            unit=self.unit,
            algo=self.algo,
            sep=SEPARATOR_LOOKUP_COMPLEX_VALUE
        )

        url = self.base_publisher_url[:] + '?{}={}'.format(
            GEO_DISTANCE_ORDERING_PARAM,
            __params
        )

        data = {}
        response = self.client.get(url, data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        # Should contain only 6 results
        self.assertEqual(len(response.data['results']), self.geo_in_count + 1)
        item_count = len(response.data['results'])

        for counter, item in enumerate(response.data['results']):
            if (counter > 1) and (counter < item_count + 1):
                self.assertLess(
                    response.data['results'][counter-1]['location']['lat'],
                    response.data['results'][counter]['location']['lat']
                )


if __name__ == '__main__':
    unittest.main()



# coding: utf-8
"""
Test functional suggestions backend.
"""

from __future__ import absolute_import, unicode_literals

import unittest

from django.core.management import call_command
from django.urls import reverse

import pytest

from rest_framework import status

import factories

from .base import BaseRestFrameworkTestCase
from .data_mixins import AddressesMixin

__title__ = 'django_elasticsearch_dsl_drf.tests.test_suggesters'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'TestFunctionalSuggesters',
)


@pytest.mark.django_db
class TestFunctionalSuggesters(BaseRestFrameworkTestCase, AddressesMixin):
    """Test functional suggesters."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        """Set up class."""
        super(TestFunctionalSuggesters, cls).setUpClass()
        cls.publishers = []
        cls.publishers.append(
            factories.PublisherFactory(
                name='Addison–Wesley',
                city='Brighton & Hove',
                state_province='East Midlands',
                country='Armenia',
            )
        )
        cls.publishers.append(
            factories.PublisherFactory(
                name='Adis International',
                city='Bristol',
                state_province='East of England',
                country='Argentina',
            )
        )
        cls.publishers.append(
            factories.PublisherFactory(
                name='Atlantic Books',
                city='Cardiff',
                state_province='North East',
                country='Belgium',
            )
        )
        cls.publishers.append(
            factories.PublisherFactory(
                name='Atlas Press',
                city='Carlisle',
                state_province='North West',
                country='Belarus',
            )
        )
        cls.publishers.append(
            factories.PublisherFactory(
                name='Book League of America',
                city='Chelmsford',
                state_province='South East',
                country='Burkina Faso',
            )
        )
        cls.publishers.append(
            factories.PublisherFactory(
                name='Strip Books of America',
                city='Chelmsford',
                state_province='South East',
                country='Burkina Faso',
            )
        )
        cls.publishers.append(
            factories.PublisherFactory(
                name='Book Works',
                city='Chester',
                state_province='South West',
                country='Burundi',
            )
        )
        cls.publishers.append(
            factories.PublisherFactory(
                name='Booktrope',
                city='Chichester',
                state_province='West Midlands',
                country='Netherlands',
            )
        )

        cls.publishers_url = reverse(
            'publisherdocument-functional-suggest',
            kwargs={}
        )

        cls.books = []
        cls.books.append(
            factories.BookFactory(
                title='Aaaaa Bbbb',
                summary='`Twas brillig, and the slithy toves '
                        'Did gyre and gimble in the wabe. '
                        'All mimsy were the borogoves '
                        'And the mome raths outgrabe.',
                publisher__name='Antares',
                publisher__country='Armenia',
            )
        )
        cls.books.append(
            factories.BookFactory(
                title='Aaaaa Cccc',
                summary='"Beware the Jabberwock, my son! '
                        'The jaws that bite, the claws that catch! '
                        'Beware the Jubjub bird, and shun '
                        'The frumious Bandersnatch!',
                publisher__name='Antares',
                publisher__country='Armenia',
            )
        )
        cls.books.append(
            factories.BookFactory(
                title='Aaaaa Dddd',
                summary='He took his vorpal sword in his hand,'
                        'Long time the manxome foe he sought --'
                        'So rested he by the Tumtum tree,'
                        'And stood awhile in thought.',
                publisher__name='Antares',
                publisher__country='Armenia',
            )
        )
        cls.books.append(
            factories.BookFactory(
                title='Eeeee Ffff',
                summary='He took his vorpal sword in his hand,'
                        'Long time the manxome foe he sought --'
                        'So rested he by the Tumtum tree,'
                        'And stood awhile in thought.',
                publisher__name='Strip Books of America',
                publisher__country='Burkina Faso',
            )
        )

        cls.books += factories.BookFactory.create_batch(
            10,
            publisher__name='Oxford University Press',
            publisher__city='Yerevan',
            publisher__state_province='Ararat',
            publisher__country='Ireland',
        )

        cls.books_url = reverse(
            'bookdocument_functional_suggester-functional-suggest',
            kwargs={}
        )

        cls.authors = []
        cls.authors.append(
            factories.AuthorFactory(
                name='John Doe',
                salutation='Aaa Bbb',
            )
        )
        cls.authors.append(
            factories.AuthorFactory(
                name='Jane Doe',
                salutation='Aaa Ccc',
            )
        )
        cls.authors.append(
            factories.AuthorFactory(
                name='Armen Doe',
                salutation='Bbb Ccc',
            )
        )

        cls.authors_url = reverse(
            'authordocument-functional-suggest',
            kwargs={}
        )

        cls.created_addresses()

        cls.sleep()
        call_command('search_index', '--rebuild', '-f')

    def _test_suggesters(self, test_data, url):
        """Test suggesters."""
        self.authenticate()

        data = {}

        for _suggester_field, _test_cases in test_data.items():

            for _test_case, _expected_results in _test_cases.items():
                # Check if response now is valid
                response = self.client.get(
                    url + '?' + _suggester_field + '=' + _test_case,
                    data
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertIn(_suggester_field, response.data)
                _unique_options = list(set([
                    __o['text']
                    for __o
                    in response.data[_suggester_field][0]['options']
                ]))
                self.assertEqual(
                    len(_unique_options),
                    len(_expected_results),
                    (_test_case, _expected_results)
                )
                self.assertEqual(
                    sorted(_unique_options),
                    sorted(_expected_results),
                    (_test_case, _expected_results)
                )

    def test_suggesters_completion(self):
        """Test suggesters completion."""
        # Testing publishers
        test_data = {
            'name_suggest__completion_prefix': {
                'Ad': ['Addison–Wesley', 'Adis International'],
                'Atl': ['Atlantic Books', 'Atlas Press'],
                'Boo': ['Book League of America', 'Book Works', 'Booktrope'],
                'Stri': ['Strip Books of America'],
            },
            'name_match_suggest__completion_match': {
                'Strip': ['Strip Books of America', ],
                'America': [
                    'Strip Books of America',
                    'Book League of America',
                ],
            },
            'country_suggest__completion_prefix': {
                'Arm': ['Armenia'],
                'Ar': ['Armenia', 'Argentina'],
                'Bel': ['Belgium', 'Belarus'],
                'Bur': ['Burkina Faso', 'Burundi'],
                'Net': ['Netherlands'],
                'Fra': [],
            }
        }
        # Testing default suggesters as well
        test_data.update(
            {
                'name_suggest': test_data['name_suggest__completion_prefix'],
                'name_match_suggest':
                    test_data['name_match_suggest__completion_match'],
                'country_suggest':
                    test_data['country_suggest__completion_prefix'],
            }
        )
        self._test_suggesters(test_data, self.publishers_url)

        # Testing books
        test_data = {
            'title_suggest_prefix__completion_prefix': {
                'Aaa': ['Aaaaa Bbbb', 'Aaaaa Cccc', 'Aaaaa Dddd'],
                'Bbb': [],
            },
        }
        # Testing default suggesters as well
        test_data.update(
            {
                'title_suggest_prefix':
                    test_data['title_suggest_prefix__completion_prefix'],
                'title.raw__completion_prefix':
                    test_data['title_suggest_prefix__completion_prefix'],
                'title_simple__completion_prefix':
                    test_data['title_suggest_prefix__completion_prefix'],
            }
        )
        self._test_suggesters(test_data, self.books_url)

        # Testing authors
        test_data = {
            'salutation_suggest__completion_prefix': {
                'Aaa': ['Aaa Bbb', 'Aaa Ccc'],
                'Bbb': ['Bbb Ccc'],
                'Hhh': [],
            },
        }
        # Testing default suggesters as well
        test_data.update(
            {
                'salutation_suggest':
                    test_data['salutation_suggest__completion_prefix'],
                'salutation.raw':
                    test_data['salutation_suggest__completion_prefix'],
            }
        )
        self._test_suggesters(test_data, self.authors_url)

    def test_suggesters_completion_no_args_provided(self):
        """Test suggesters completion with no args provided."""
        data = {}
        # Check if response now is valid
        response = self.client.get(self.publishers_url, data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    # def test_suggesters_term(self):
    #     """Test suggesters term."""
    #     # Testing books
    #     test_data = {
    #         'summary_suggest__term': {
    #             'borogovse': ['borogov'],
    #             'Tumtus': ['tumtum'],
    #             'Jabberwok': ['jabberwock'],
    #             'tovse': ['tove', 'took', 'twas'],
    #         },
    #     }
    #     self._test_suggesters(test_data, self.books_url)
    #
    # def test_suggesters_phrase(self):
    #     """Test suggesters phrase."""
    #     # Testing books
    #     test_data = {
    #         'summary_suggest__phrase': {
    #             'slith tovs': ['slithi tov'],
    #             'mimsy boroto': ['mimsi borogov'],
    #         },
    #     }
    #     self._test_suggesters(test_data, self.books_url)
    #
    # def test_nested_fields_suggesters_completion(self):
    #     """Test suggesters completion for nested fields."""
    #     # Testing cities and countries
    #     test_data = {
    #         'city_suggest__completion': {
    #             'Ye': ['Yerevan', 'Yeovil'],
    #             'Yer': ['Yerevan'],
    #             'Ams': ['Amsterdam'],
    #             'Du': ['Dublin'],
    #             'Ne': [],
    #         },
    #         'country_suggest__completion': {
    #             'Arm': ['Armenia'],
    #             'Ar': ['Armenia', 'Argentina'],
    #             'Re': ['Republic of Ireland'],
    #             'Net': ['Netherlands'],
    #             'Fra': [],
    #         }
    #     }
    #     self._test_suggesters(test_data, self.addresses_suggest_url)


if __name__ == '__main__':
    unittest.main()



"""
Test faceted search backend.
"""

from __future__ import absolute_import

import unittest

from django.core.management import call_command
from django.urls import reverse

import pytest

from rest_framework import status

from books import constants
import factories

from .base import BaseRestFrameworkTestCase

__title__ = 'django_elasticsearch_dsl_drf.tests.test_faceted_search'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'TestFacetedFilteredSearch',
)


@pytest.mark.django_db
class TestFacetedFilteredSearch(BaseRestFrameworkTestCase):
    """Test faceted filtered search."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        super(TestFacetedFilteredSearch, cls).setUpClass()

        # published with a known title
        cls.prefixed = factories.BookFactory.create(
            **{
                'title': 'My first book',
                'state': constants.BOOK_PUBLISHING_STATUS_PUBLISHED,
            }
        )

        cls.published_count = 11
        cls.published = factories.BookFactory.create_batch(
            cls.published_count - 1,
            **{
                'state': constants.BOOK_PUBLISHING_STATUS_PUBLISHED,
            }
        )

        cls.not_published_count = 10
        cls.not_published = factories.BookFactory.create_batch(
            cls.not_published_count,
            **{
                'state': constants.BOOK_PUBLISHING_STATUS_NOT_PUBLISHED,
            }
        )

        cls.all_count = cls.published_count + cls.not_published_count

        cls.sleep()
        call_command('search_index', '--rebuild', '-f')

    def test_list_results_no_facets(self):
        """List results without facets."""
        self.authenticate()
        url = reverse('bookdocument_faceted_filtered-list')

        # Make request
        no_args_response = self.client.get(url)
        self.assertEqual(no_args_response.status_code, status.HTTP_200_OK)

        # Should contain `self.all_count` results
        self.assertEqual(len(no_args_response.data['results']), self.all_count)

        # Should contain no facets
        print(no_args_response.data['facets'])
        self.assertEqual(no_args_response.data['facets'], {})

    def test_list_results_facet_no_filter(self):
        self.authenticate()
        url = reverse('bookdocument_faceted_filtered-list')

        response = self.client.get(url + '?facet=state')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # Should contain `self.all_count` results
        self.assertEqual(len(response.data['results']), self.all_count)

        # Should contain 1 facet
        self.assertEqual(len(response.data['facets']), 1)
        self.assertEqual(response.data['facets']['_filter_state']['state']['buckets'], [{
            "doc_count": self.published_count,
            "key": "published"
        }, {
            "doc_count": self.not_published_count,
            "key": "not_published"
        }])

    def test_list_results_facet_and_filter(self):
        self.authenticate()
        url = reverse('bookdocument_faceted_filtered-list')

        response = self.client.get(url + '?facet=state&state={}'.format(constants.BOOK_PUBLISHING_STATUS_PUBLISHED))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # Should contain all published books
        self.assertEqual(len(response.data['results']), self.published_count)

        # Should contain 1 facet
        self.assertEqual(len(response.data['facets']), 1)
        self.assertEqual(response.data['facets']['_filter_state']['state']['buckets'], [{
            "doc_count": self.published_count,
            "key": "published"
        }, {
            "doc_count": self.not_published_count,
            "key": "not_published"
        }])

    def test_list_results_facet_and_filter_on_non_faceted_field(self):
        self.authenticate()
        url = reverse('bookdocument_faceted_filtered-list')

        response = self.client.get(url + '?facet=state&state={}&title={}'.format(
            constants.BOOK_PUBLISHING_STATUS_PUBLISHED,
            'My first book',
        ))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # Should contain 1 result
        self.assertEqual(len(response.data['results']), 1)

        # Should contain 1 facet which matches only the single book, since that is applied as a pre-filter
        self.assertEqual(len(response.data['facets']), 1)
        self.assertEqual(response.data['facets']['_filter_state']['state']['buckets'], [{
            "doc_count": 1,
            "key": "published"
        }])

    def test_list_results_facet_and_filter_only_on_non_faceted_field(self):
        self.authenticate()
        url = reverse('bookdocument_faceted_filtered-list')

        response = self.client.get(url + '?facet=state&title={}'.format('My first book'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # Should contain 1 result
        self.assertEqual(len(response.data['results']), 1)

        # Should contain 1 facet which matches only the single book, since that is applied as a pre-filter
        self.assertEqual(len(response.data['facets']), 1)
        self.assertEqual(response.data['facets']['_filter_state']['state']['buckets'], [{
            "doc_count": 1,
            "key": "published"
        }])


if __name__ == '__main__':
    unittest.main()



"""
Test source backend.
"""

from __future__ import absolute_import

import unittest

from django.core.management import call_command
from django.urls import reverse

import pytest

from rest_framework import status

import factories

from .base import BaseRestFrameworkTestCase

__title__ = 'django_elasticsearch_dsl_drf.tests.test_source'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'TestSource',
)


@pytest.mark.django_db
class TestSource(BaseRestFrameworkTestCase):
    """Test source."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        super(TestSource, cls).setUpClass()

        cls.books_count = 10
        cls.books = factories.BookFactory.create_batch(
            cls.books_count,
        )

        cls.special_books_count = 10
        cls.special_books = factories.BookFactory.create_batch(
            cls.books_count,
            title='Twenty Thousand Leagues Under the Sea',
            description="""
            The title refers to the distance traveled while under the sea and
            not to a depth, as twenty thousand leagues is over six times the
            diameter, and nearly twice the circumference of the Earth
            """,
            summary="""
            Margaret Drabble argues that Twenty Thousand Leagues Under the Sea
            anticipated the ecology movement and shaped the French avant-garde
            """,
        )
        cls.all_books_count = cls.special_books_count + cls.books_count

        cls.sleep()
        call_command('search_index', '--rebuild', '-f')

    def _list_results(self):
        """List results."""
        self.authenticate()

        url = reverse('bookdocument_source-list', kwargs={}) + '?search=twenty'

        # Make request
        response = self.client.get(url, {})
        self.assertEqual(
            response.status_code,
            status.HTTP_200_OK
        )

        # Should contain 10 results
        self.assertEqual(
            len(response.data['results']),
            self.special_books_count
        )

        expected_keys = {'id', 'title'}
        # Should only contain 'id' and 'title'.
        for result in response.data['results']:
            self.assertEqual(
                expected_keys,
                set(result.keys())
            )

    def test_list_results(self):
        """Test list results."""
        return self._list_results()


if __name__ == '__main__':
    unittest.main()



"""Compound search backend."""

from .base import BaseSearchFilterBackend
from .query_backends import (
    MatchQueryBackend,
    NestedQueryBackend,
)

__title__ = 'django_elasticsearch_dsl_drf.filter_backends.search.compound'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'CompoundSearchFilterBackend',
)


class CompoundSearchFilterBackend(BaseSearchFilterBackend):
    """Compound search backend."""

    query_backends = [
        MatchQueryBackend,
        NestedQueryBackend,
    ]



"""
Pip helpers module.
"""
import subprocess
import sys

__title__ = 'django_elasticsearch_dsl_drf.tests.pip_helpers'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'check_if_installed',
    'get_installed_packages',
)


def get_installed_packages(with_versions=False):
    """Get installed packages.

    :param with_versions: If set to True, returned with versions.
    :type with_versions: bool
    :return:
    :rtype: list
    """
    reqs = subprocess.check_output([sys.executable, '-m', 'pip', 'freeze'])
    if with_versions:
        return set([tuple(r.decode().split('==')) for r in reqs.split()])
    else:
        return set([r.decode().split('==')[0] for r in reqs.split()])


def check_if_installed(package, installed_packages=None):
    """Check if package is installed.

    :param package:
    :param installed_packages:
    :type package: str
    :type installed_packages: iterable
    :return:
    :rtype: bool
    """
    if installed_packages is None:
        installed_packages = get_installed_packages(with_versions=False)
    return package in installed_packages



__title__ = 'django_elasticsearch_dsl_drf.filter_backends.search.' \
            'query_backends.base'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2017-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = ('BaseSearchQueryBackend',)


class BaseSearchQueryBackend(object):
    """Search query backend."""

    @classmethod
    def construct_search(cls, request, view, search_backend):
        """Construct search.

        :param request:
        :param view:
        :param search_backend:
        :return:
        """
        raise NotImplementedError(
            "You should implement `construct_search` method in your {} class"
            "".format(cls.__name__)
        )
